#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import datetime
from collections import defaultdict
from typing import Iterable, Type, TypeVar

from bson import ObjectId
from pymongo.collection import Collection


class MongoFieldSchemator:
    _bsontype_of_pytypes = {
        str: 'string',
        dict: 'object',
        type(None): 'null',
        datetime.datetime: 'date',
        int: 'long',
        bool: 'bool',
        ObjectId: 'objectId',
    }

    _max_enum_length = 7
    _max_type_length = 3

    def __init__(self):
        self._bsontypes = set()
        self._enum_values = set()

    @property
    def _enum_enabled(self) -> bool:
        if len(self._enum_values) > self._max_enum_length:
            return False
        return set not in self._enum_values

    def _add_bsontype(self, val):
        try:
            bsontype = self._bsontype_of_pytypes[type(val)]
        except KeyError:
            return
        self._bsontypes.add(bsontype)

    def _add_enum_value(self, val):
        if self._enum_enabled:
            return
        try:
            self._enum_values.add(val)
        except TypeError:
            self._enum_values.add(set)

    def add(self, val):
        self._add_enum_value(val)
        self._add_bsontype(val)

    @property
    def bsontype(self) -> str | list[str] | None:
        if not self._bsontypes:
            return
        if len(self._bsontypes) > self._max_type_length:
            return
        _bsontypes = list(self._bsontypes)
        if len(_bsontypes) > 1:
            return _bsontypes
        return _bsontypes[0]

    @property
    def enum(self) -> list | None:
        if not self._enum_enabled or not self._enum_values:
            return
        return list(sorted(self._enum_values))

    def to_jsonschema_property(self) -> dict:
        """
        Returns:
            e.g.
            {
               "bsonType": "string",
               "enum": ["low", "mid", "high"],
            }
        """
        p = {'bsonType': self.bsontype, 'enum': self.enum}
        return {k: v for k, v in p.items() if v is not None}


T = TypeVar('T')


class MongoDocumentSchemator:
    def __init__(self, fieldnames: set[str]):
        if not isinstance(fieldnames, set):
            fieldnames = set(fieldnames)
        self._fieldnames = fieldnames
        self._required = fieldnames.copy()
        self._fields = defaultdict(MongoFieldSchemator)

    @classmethod
    def from_initial_records(cls: Type[T], records: Iterable[dict]) -> T:
        fieldnames = set()
        for rec in records:
            fieldnames.update(set(rec))
        return cls(fieldnames)

    @classmethod
    def from_collection(cls: Type[T], coll: Collection) -> T:
        skmtr = MongoDocumentSchemator.from_initial_records(coll.find())
        skmtr.add_many(coll.find())
        return skmtr

    @property
    def fieldnames(self) -> set:
        return self._fieldnames.copy()

    def add(self, record: dict):
        for key in self._fieldnames:
            val = record.get(key)
            self._fields[key].add(val)
            if val is None and val in self._required:
                self._required.remove(key)

    def add_many(self, records: Iterable[dict]):
        for rec in records:
            self.add(rec)

    def get_properties(self) -> dict[str, dict]:
        return {k: v.to_jsonschema_property() for k, v in self._fields.items()}

    def get_required(self) -> list[str]:
        return list(sorted(self._required))

    def to_jsonschema(self) -> dict:
        return {
            # $jsonSchema keyword '$schema' is not currently supported
            # "$schema": "http://json-schema.org/draft-04/schema#",
            'type': 'object',
            'properties': self.get_properties(),
            'required': self.get_required(),
        }


__all__ = ['MongoFieldSchemator', 'MongoDocumentSchemator']



#!/usr/bin/env python3
# coding: utf-8

from pymongo.collection import Collection


def kv_load(coll: Collection, key: str):
    doc = coll.find_one({'_id': key}, projection={'_id': False})
    if doc is None:
        return
    try:
        return doc['_']
    except KeyError:
        return doc


def kv_save(coll: Collection, key: str, val):
    filtr = {'_id': key}
    # explode dict if '_' and '_id' are not in it -- be less nested
    if isinstance(val, dict) and '_' not in val and '_id' not in val:
        replacement = val
    else:
        replacement = {'_': val}
    return coll.replace_one(filtr, replacement, upsert=True)



#!/usr/bin/env python3
# coding: utf-8

import re
import json

from bson.json_util import loads
import datetime


def fix_mongoshell_json(s: str):
    # https://stackoverflow.com/a/11886476/2925169
    s = re.sub(
        r':\s*ObjectId\s*\(\s*\"(\S+)\"\s*\)',
        r': {"$oid": "\1"}', s
    )
    s = re.sub(
        r':\s*ISODate\s*\(\s*(\S+)\s*\)',
        r': {"$date": \1}', s
    )
    return s


def loads_mongoshell_json(s: str):
    return loads(fix_mongoshell_json(s))


def inplace_fix_mongoshell_jsonfile(path: str) -> bool:
    now = datetime.datetime.now()
    raw = open(path).read()
    fixed = fix_mongoshell_json(raw)
    if raw == fixed:
        return False
    try:
        json.loads(fixed)
    except json.JSONDecodeError:
        return False
    with open(path + f'.{now:%Y%m%d-%H%M%S}.bak', 'w') as fout:
        fout.write(raw)
    with open(path, 'w') as fout:
        fout.write(fixed)
    return True


def main():
    import sys
    inplace_fix_mongoshell_jsonfile(sys.argv[1])


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# coding: utf-8

import argparse

from volkanic.cmdline import CommandOptionDict


def _add_hyphens(parts: list) -> list:
    if not parts:
        return parts
    key = str(parts[0])
    if not key.startswith('-'):
        prefix = '--' if len(key) > 1 else '-'
        key = prefix + key
    return [key] + parts[1:]


class CommandOptionDictMongoish(CommandOptionDict):
    @classmethod
    def _explode(cls, key, val):
        parts = super()._explode(key, val)
        parts = _add_hyphens(parts)
        if len(parts) != 2:
            return parts
        key, val = parts
        if key.startswith('--'):
            return [f'{key}={val}']


class ArgumentParserMongoish(argparse.ArgumentParser):
    def add_argument_mongoi_host(self):
        self.add_argument('-H', '--host', help='host as in MongoInterface')

    def add_arguments_host_port(self, hostname='127.0.0.1'):
        add = self.add_argument
        add('-H', '--hostname', default=hostname, help='hostname')
        add('-p', '--port', type=int, default=27017, help='port number')

    def add_argument_db_name(self):
        self.add_argument('-d', '--db-name', help='database name')

    def add_argument_coll_name(self):
        self.add_argument('-c', '--coll-name', help='collection name')



#!/usr/bin/env python3
# coding: utf-8

from pymongo.collection import Collection


def _namemap_to_project(namemap: dict):
    project = {}
    for new_name, old_name in namemap.items():
        if not old_name.startswith('$'):
            old_name = '$' + old_name
        project[new_name] = {'$ifNull': [old_name, None]}
    return project


def _namemap_from_fieldlist(fieldlist: list):
    return {k.split('.')[-1]: k for k in fieldlist}


def find_with_renaming(
        coll: Collection, filtr: dict, namemap: dict, sort: dict = None):
    pipelines = [
        {'$match': filtr},
        {'$sort': sort or {'_id': -1}},
        {'$project': _namemap_to_project(namemap)},
    ]
    return coll.aggregate(pipelines)


def find_one_with_renaming(
        coll: Collection, filtr: dict, namemap: dict, sort: dict = None):
    pipelines = [
        {'$match': filtr},
        {'$sort': sort or {'_id': -1}},
        {'$limit': 1},
        {'$project': _namemap_to_project(namemap)},
    ]
    docs = list(coll.aggregate(pipelines))
    if docs:
        return docs[0]



#!/usr/bin/env python3
# coding: utf-8


from typing import List
from joker.mongodb.tools.cmdline import CommandOptionDictMongoish

import re
from pymongo.database import Database
from joker.mongodb import utils


def find_excluding_coll_names(db: Database, regexes: List[str]) -> list:
    for coll_name in db.list_collection_names():
        ns = f'{db.name}.{coll_name}'
        for pat in regexes:
            if re.search(pat, ns):
                yield coll_name


def dump(params: dict):
    cod = CommandOptionDictMongoish(params)
    cod.executable = 'mongodump'
    cod.run(dry=True)


def smart_dump(db: Database, outpath: str, excl_regexes: List[str]):
    excl_coll_names = find_excluding_coll_names(db, excl_regexes)
    params = {
        'gzip': True, 'out': outpath, 'db': db.name,
        'excludeCollection': tuple(excl_coll_names),
    }
    cfg = utils.infer_params(db.client)
    keys = ['host', 'port', 'username', 'password']
    for key in keys:
        val = cfg.get(key)
        if val:
            params[key] = val
    return params
    # dump(params)



#!/usr/bin/env python3
# coding: utf-8

import logging
import time
from collections import UserDict

import pymongo
from bson import Timestamp, json_util, ObjectId
from pymongo import MongoClient

_logger = logging.getLogger(__name__)


class OplogRecord(UserDict):
    @property
    def op(self) -> str:
        return self.data.get('op')

    @property
    def ns(self) -> str:
        return self.data.get('ns')

    @property
    def ts(self) -> Timestamp:
        return self.data.get('ts')

    def to_mongo_doc(self):
        doc = self.data.copy()
        doc['ns'] = self.ns
        if 'o' in doc:
            doc['o'] = json_util.dumps(doc['o'])
        return doc

    @classmethod
    def from_mongo_doc(cls, doc: dict):
        doc = dict(doc)
        if 'o' in doc:
            doc['o'] = json_util.loads(doc['o'])
        return cls(doc)

    @property
    def upstream_id(self) -> str:
        try:
            return str(self.data['o']['_id'])
        except (KeyError, TypeError):
            pass
        try:
            return str(self.data['o2']['_id'])
        except (KeyError, TypeError):
            pass

    @property
    def id_(self) -> ObjectId:
        if id_ := self.data.get('_id'):
            return ObjectId(id_)


# Caution: do NOT run multiple threads / processes of this!!
# TODO: consider thread safety
class OplogTailer(object):
    _ns_exclude = {}
    _db_exclude = {'config', 'local', 'admin'}
    record_cls = OplogRecord

    def __init__(
            self, upstream_client: MongoClient, ts: Timestamp,
            ns_pattern: str = None, ns_exclude: str = None):
        """
        Args:
            upstream_client:
            ts: starting timestamp (seconds since epoch), e.g. 1642477123
            ns_pattern: regex
            ns_exclude: regex
        """
        db = upstream_client.get_database('local')
        self.oplog_coll = db.get_collection('oplog.rs')
        self.ts = ts
        self.ns_pattern = ns_pattern
        self.ns_exclude = ns_exclude
        self._cursor = self._get_cursor()

    def _get_where_clause(self):
        tests = ['this.op != "n"']
        if self.ns_pattern is not None:
            tests.append('this.ns.match(/{}/)'.format(self.ns_pattern))
            # tests.append(f'this.ns.match(/{self.ns_pattern}/)')
        if self.ns_exclude is not None:
            tests.append('!this.ns.match(/{}/)'.format(self.ns_exclude))
            # tests.append(f'!this.ns.match(/{self.ns_exclude}/)')
        return '&&'.join(tests)

    def _get_cursor(self):
        filtr = {'ts': {'$gt': self.ts}}
        # https://pymongo.readthedocs.io/en/stable/examples/tailable.html
        # TODO: consider using $expr for mongo >= 3.6
        _logger.info('OplogTailer._get_cursor, filtr=%s', filtr)
        cursor = self.oplog_coll.find(
            filtr, oplog_replay=True,
            cursor_type=pymongo.CursorType.TAILABLE_AWAIT,
        )
        where_clause = self._get_where_clause()
        if where_clause:
            cursor = cursor.where(where_clause)
        return cursor

    def _reset_cursor(self):
        self._cursor.close()
        self._cursor = self._get_cursor()

    def _check_ns(self, doc: dict):
        ns = doc.get('ns')
        if not ns:
            return False
        if ns in self._ns_exclude:
            return False
        db_name = ns.split('.')[0]
        if db_name in self._db_exclude:
            return False
        return True

    def _fetch_next(self):
        try:
            return self._cursor.next()
        except (StopIteration, ConnectionResetError):
            self._reset_cursor()

    def __iter__(self):
        return self

    def __next__(self) -> 'record_cls':
        while True:
            doc = self._fetch_next()
            if doc is None:
                time.sleep(1)
                continue
            if not self._check_ns(doc):
                continue
            self.ts = doc.get('ts')
            return self.record_cls(doc)



#!/usr/bin/env python3
# coding: utf-8

"""
This module is DEPRECATED.
"""

from gridfs import GridFS
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from joker.mongodb import utils
from joker.mongodb.interfaces import CollectionInterface


class DatabaseInterface:
    def __init__(self, db: Database):
        self._db = db

    def inspect_storage_sizes(self):
        return utils.inspect_mongo_storage_sizes(self._db)

    def print_storage_sizes(self):
        return utils.print_mongo_storage_sizes(self._db)


class MongoClientExtended(MongoClient):
    """An extended client-side representation of a mongodb cluster."""

    def __repr__(self):
        cn = self.__class__.__name__
        return "{}({})".format(cn, self._repr_helper())

    def inspect_storage_sizes(self):
        return utils.inspect_mongo_storage_sizes(self)

    def print_storage_sizes(self):
        return utils.print_mongo_storage_sizes(self)

    def get_db(self, db_name: str) -> Database:
        return self.get_database(db_name)

    def get_dbi(self, db_name: str) -> DatabaseInterface:
        return DatabaseInterface(self.get_database(db_name))

    def get_coll(self, db_name: str, coll_name: str) -> Collection:
        db = self.get_database(db_name)
        return db.get_collection(coll_name)

    def get_ci(self, db_name: str, coll_name: str) -> CollectionInterface:
        coll = self.get_coll(db_name, coll_name)
        return CollectionInterface(coll)

    def get_gridfs(self, db_name: str, coll_name: str = 'fs') \
            -> GridFS:
        # avoid names like "images.files.files"
        if coll_name.endswith('.files') or coll_name.endswith('.chunks'):
            coll_name = coll_name.rsplit('.', 1)[0]
        db = self.get_database(db_name)
        return GridFS(db, collection=coll_name)


#!/usr/bin/env python3
# coding: utf-8

from gridfs import GridFS


def _get_gridfs_namespace(fs: GridFS):
    coll = getattr(fs, '_GridFS__files')
    db = coll.database
    return f'{db.name}.{coll.name}'




#!/usr/bin/env python3
# coding: utf-8

import datetime

import pymongo.errors
from pymongo import ReturnDocument
from pymongo.collection import Collection


class TransactionHelper:
    def __init__(self, coll: Collection):
        self.coll = coll

    def lock(self, name: str, ttl: int = 60):
        now = datetime.datetime.now()
        expire_at = now + datetime.timedelta(seconds=ttl)
        self.coll.delete_many(
            {'expire_at': {'$lt': now}, 'type': 'lock'},
        )
        record = {
            '_id': name,
            'type': 'lock',
            'expire_at': expire_at,
        }
        try:
            self.coll.insert_one(record)
        except pymongo.errors.DuplicateKeyError:
            return False
        return True

    def unlock(self, name: str):
        self.coll.delete_one({'_id': name, 'type': 'lock'})

    def get_next_serial_code(self, prefix='ID', length=6):
        doc = self.coll.find_one_and_update(
            {'_id': prefix, 'type': 'serial'},
            {'$inc': {'i': 1}},
            upsert=True, return_document=ReturnDocument.AFTER,
        )
        return prefix + str(doc['i']).zfill(length)



#!/usr/bin/env python3
# coding: utf-8

from collections import defaultdict
from typing import Union

import pymongo.errors
from bson import ObjectId, json_util
from gridfs import GridFS
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.database import Database
from volkanic.utils import printerr

from joker.mongodb import utils
from joker.mongodb.tools import kvstore


class CollectionInterface:
    def __init__(self, coll: Collection, filtr=None, projection=None):
        self._coll = coll
        self.filtr = filtr or {}
        self.projection = projection

    def exist(self, filtr: Union[ObjectId, dict]):
        return self._coll.find_one(filtr, projection=[])

    def kv_load(self, key: str):
        return kvstore.kv_load(self._coll, key)

    def kv_save(self, key: str, val):
        return kvstore.kv_save(self._coll, key, val)

    def find_recent_by_count(self, count=50) -> Cursor:
        cursor = self._coll.find(self.filtr, projection=self.projection)
        return cursor.sort([('_id', -1)]).limit(count)

    def find_most_recent_one(self) -> dict:
        recs = list(self.find_recent_by_count(1))
        if recs:
            return recs[0]

    def _insert(self, records):
        if records:
            self._coll.insert_many(records, ordered=False)

    @staticmethod
    def _check_for_uniqueness(records, uk):
        vals = [r.get(uk) for r in records]
        uniq_vals = set(vals)
        if len(vals) != len(uniq_vals):
            raise ValueError('records contain duplicating keys')

    def make_fusion_record(self):
        fusion_record = {}
        contiguous_stale_count = -1
        for skip in range(1000):
            record = self._coll.find_one(sort=[('$natural', -1)], skip=skip)
            if not record:
                continue
            contiguous_stale_count += 1
            for key, val in record.items():
                if not record.get(key):
                    fusion_record[key] = val
                    contiguous_stale_count = -1
            if contiguous_stale_count > 10:
                return fusion_record
        return fusion_record

    def query_uniq_values(self, fields: list, limit=1000):
        latest = [('_id', -1)]
        records = self._coll.find(sort=latest, projection=fields, limit=limit)
        uniq = defaultdict(set)
        for key in fields:
            for rec in records:
                val = rec.get(key)
                uniq[key].add(val)
        return uniq


class MongoInterface:
    """A interface for multiple mongodb clusters."""

    def __init__(
            self, hosts: dict,
            default: str = 'localhost.default',
            aliases: dict = None):
        self.default_host, self.default_db_name = default.split('.')
        self.hosts = hosts
        self.aliases = aliases or {}
        self._clients = {}

    @classmethod
    def from_config(cls, options: dict):
        params = {
            'default': options.pop('_default', None),
            'aliases': options.pop('_aliases', None),
        }
        return cls(options, **params)

    def get_mongo(self, host: str = None) -> MongoClient:
        if host is None:
            host = self.default_host
        try:
            return self._clients[host]
        except KeyError:
            pass
        # host pass through as MongoClient argument
        params = self.hosts.get(host, host)
        if isinstance(params, str):
            params = {'host': params}
        return self._clients.setdefault(host, MongoClient(**params))

    @property
    def db(self) -> Database:
        return self.get_db(self.default_host, self.default_db_name)

    def _check_coll_triple(self, names: tuple) -> tuple:
        n = len(names)
        if n == 1:
            return self.default_host, self.default_db_name, names[0]
        elif n == 3:
            return names
        else:
            c = self.__class__.__name__
            msg = 'requires 1 or 3 arguments, got {}'.format(c, n)
            raise ValueError(msg)

    def __call__(self, *names) -> Collection:
        names = self._check_coll_triple(names)
        return self.get_coll(*names)

    def get_db(self, host: str, db_name: str) -> Database:
        mongo = self.get_mongo(host)
        db_name = self.aliases.get(db_name, db_name)
        return mongo.get_database(db_name)

    def get_coll(self, host: str, db_name: str, coll_name: str) \
            -> Collection:
        db = self.get_db(host, db_name)
        return db.get_collection(coll_name)

    def get_gridfs(self, host: str, db_name: str, coll_name: str = 'fs') \
            -> GridFS:
        assert not coll_name.endswith('.files')
        assert not coll_name.endswith('.chunks')
        db = self.get_db(host, db_name)
        return GridFS(db, collection=coll_name)


class MongoInterfaceExtended(MongoInterface):
    def _get_target(self, host: str, db_name: str = None):
        if db_name is None:
            return self.get_mongo(host)
        return self.get_db(host, db_name)

    def inspect_storage_sizes(self, host: str, db_name: str = None):
        target = self._get_target(host, db_name)
        return utils.inspect_mongo_storage_sizes(target)

    def print_storage_sizes(self, host: str, db_name: str = None):
        target = self._get_target(host, db_name)
        return utils.print_mongo_storage_sizes(target)

    def get_ci(self, *names, **kwargs) -> CollectionInterface:
        coll = self.__call__(*names)
        return CollectionInterface(coll, **kwargs)

    # TODO: support BSON
    def restore_a_file(self, lines, inner_path: str, empty_coll_only=True):
        host, db_name, coll_name = \
            utils.infer_coll_triple_from_filename(inner_path)
        if coll_name == 'system.indexes':
            return
        coll = self.get_coll(host, db_name, coll_name)
        if empty_coll_only and coll.find_one(projection=[]):
            printerr(inner_path, 'skipped')
            return
        for ix, line in enumerate(lines):
            doc = json_util.loads(line)
            id_ = doc.get('_id', '')
            printerr(inner_path, ix, id_, '...', end=' ')
            try:
                coll.insert_one(doc)
            except pymongo.errors.DuplicateKeyError:
                printerr('DuplicateKeyError')
            else:
                printerr('completed')



#!/usr/bin/env python3
# coding: utf-8

import logging
import os.path
from typing import Union

import bson.json_util
import volkanic.utils
from joker.cast.numeric import human_filesize
from joker.textmanip.tabular import tabular_format
from pymongo import MongoClient
from pymongo.database import Database
from pymongo.monitoring import (
    CommandListener, CommandStartedEvent,
    CommandFailedEvent, CommandSucceededEvent,
)


def is_in_replset(mongo: MongoClient) -> bool:
    db = mongo.get_database('admin')
    return 'replication' in db.command('getCmdLineOpts')['parsed']


def inspect_mongo_storage_sizes(target: Union[MongoClient, Database]):
    if isinstance(target, MongoClient):
        return {r['name']: r['sizeOnDisk'] for r in target.list_databases()}
    size_of_collections = {}
    for coll_name in target.list_collection_names():
        info = target.command('collStats', coll_name)
        size_of_collections[info['ns']] = info['storageSize']
    return size_of_collections


def print_mongo_storage_sizes(target: Union[MongoClient, Database]):
    s_rows = list(inspect_mongo_storage_sizes(target).items())
    s_rows.sort(key=lambda r: r[1], reverse=True)
    rows = []
    for k, v in s_rows:
        num, unit = human_filesize(v)
        rows.append([round(num), unit, k])
    for row in tabular_format(rows):
        print(*row)


def indented_json_dumps(obj, **kwargs):
    kwargs.setdefault('dumps', bson.json_util.dumps)
    return volkanic.utils.indented_json_dumps(obj, **kwargs)


def indented_json_print(obj, **kwargs):
    kwargs.setdefault('dumps', indented_json_dumps)
    return volkanic.utils.indented_json_print(obj, **kwargs)


def infer_coll_triple_from_filename(path: str):
    """
    >>> p = "somedir/local.retail.customers.6789.json"
    >>> infer_coll_triple_from_filename(p)
    ['local', 'retail', 'customers']
    """
    filename = os.path.split(path)[1]
    coll_fullname = filename.rsplit('.', 2)[0]
    return coll_fullname.split('.', 2)


def infer_params(mongo: MongoClient):
    params = dict(zip(['host', 'port'], mongo.address))
    params.update(mongo._MongoClient__options._options)  # noqa
    return params


class MongoCommandLogger(CommandListener):
    _registered = False
    _logger = logging.getLogger('_mongodb')
    _level = logging.DEBUG

    @staticmethod
    def _fmt_opid(event):
        if event.request_id == event.operation_id:
            return event.request_id
        return '{}.{}'.format(event.request_id, event.operation_id)

    @staticmethod
    def _fmt_url(event):
        addr = ':'.join(str(s) for s in event.connection_id)
        return '{}/{}'.format(addr, event.database_name)

    def started(self, event: CommandStartedEvent):
        if not self._logger.isEnabledFor(self._level):
            return
        parts = [
            self._fmt_opid(event),
            'started',
            self._fmt_url(event),
            event.command,
        ]
        msg = " ".join(str(s) for s in parts)
        self._logger.debug(msg)

    def succeeded(self, event: CommandSucceededEvent):
        if not self._logger.isEnabledFor(self._level):
            return
        parts = [
            self._fmt_opid(event),
            'succeeded',
            int(event.duration_micros / 1000),
        ]
        msg = " ".join(str(s) for s in parts)
        self._logger.debug(msg)

    def failed(self, event: CommandFailedEvent):
        if not self._logger.isEnabledFor(self._level):
            return
        parts = [
            self._fmt_opid(event),
            'failed',
            event.duration_micros,
        ]
        msg = " ".join(str(s) for s in parts)
        self._logger.debug(msg)
